package org.eclipse.jdt.internal.compiler.lookup;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BoundSet;
import org.eclipse.jdt.internal.compiler.lookup.InferenceContext18;
import org.eclipse.jdt.internal.compiler.lookup.InferenceVariable;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBound;

class ThreeSets {
  Set<TypeBound> superBounds;
  
  Set<TypeBound> sameBounds;
  
  Set<TypeBound> subBounds;
  
  TypeBinding instantiation;
  
  Map<InferenceVariable, TypeBound> inverseBounds;
  
  public ParameterizedTypeBinding mergeTypeParameters(ParameterizedTypeBinding current, ParameterizedTypeBinding newB) {
    TypeBinding[] curTypeArgs = current.typeArguments();
    TypeBinding[] newTypeArgs = newB.typeArguments();
    TypeBinding[] merged = new TypeBinding[curTypeArgs.length];
    System.arraycopy(curTypeArgs, 0, merged, 0, curTypeArgs.length);
    boolean wasMerged = false;
    for (int i = 0; i < curTypeArgs.length; i++) {
      if (!TypeBinding.equalsEquals(curTypeArgs[i], newTypeArgs[i])) {
        if (curTypeArgs[i].isCapture() || newTypeArgs[i].isCapture())
          return null; 
        if (curTypeArgs[i] instanceof InferenceVariable) {
          if (!(newTypeArgs[i] instanceof InferenceVariable)) {
            ThreeSets three = (ThreeSets)BoundSet.this.boundsPerVariable.get(curTypeArgs[i]);
            if (three != null && three.sameBounds != null && three.sameBounds.contains(new TypeBound((InferenceVariable)curTypeArgs[i], newTypeArgs[i], 4))) {
              merged[i] = newTypeArgs[i];
              wasMerged = true;
            } 
          } else if (!curTypeArgs[i].equals(newTypeArgs[i])) {
            return null;
          } 
        } else if (!(newTypeArgs[i] instanceof InferenceVariable)) {
          if (!TypeBinding.equalsEquals(curTypeArgs[i], newTypeArgs[i]))
            return null; 
        } else {
          ThreeSets three = (ThreeSets)BoundSet.this.boundsPerVariable.get(newTypeArgs[i]);
          if (three == null || three.sameBounds == null || !three.sameBounds.contains(new TypeBound((InferenceVariable)newTypeArgs[i], curTypeArgs[i], 4)))
            return null; 
        } 
      } 
    } 
    if (wasMerged) {
      ParameterizedTypeBinding clone = (ParameterizedTypeBinding)current.clone((TypeBinding)current.enclosingType());
      clone.arguments = merged;
      return clone;
    } 
    return null;
  }
  
  public boolean addBound(TypeBound bound) {
    Iterator<TypeBound> it = null;
    switch (bound.relation) {
      case 3:
        if (this.superBounds == null)
          this.superBounds = new HashSet<>(); 
        if (CompilerOptions.useunspecdtypeinferenceperformanceoptimization && 
          !bound.right.isProperType(true)) {
          it = this.superBounds.iterator();
          while (it.hasNext()) {
            TypeBound b = it.next();
            if (bound.right.isParameterizedType() && b.right.isParameterizedType() && 
              b.right.original() == bound.right.original()) {
              ParameterizedTypeBinding parameterizedTypeBinding = mergeTypeParameters((ParameterizedTypeBinding)b.right, 
                  (ParameterizedTypeBinding)bound.right);
              if (parameterizedTypeBinding != null) {
                b.right = (TypeBinding)parameterizedTypeBinding;
                return false;
              } 
            } 
          } 
        } 
        return this.superBounds.add(bound);
      case 4:
        if (this.sameBounds == null)
          this.sameBounds = new HashSet<>(); 
        return this.sameBounds.add(bound);
      case 2:
        if (this.subBounds == null)
          this.subBounds = new HashSet<>(); 
        if (CompilerOptions.useunspecdtypeinferenceperformanceoptimization && 
          !bound.right.isProperType(true)) {
          it = this.subBounds.iterator();
          while (it.hasNext()) {
            TypeBound b = it.next();
            if (bound.right.isParameterizedType() && b.right.isParameterizedType() && 
              b.right.original() == bound.right.original()) {
              ParameterizedTypeBinding parameterizedTypeBinding = mergeTypeParameters((ParameterizedTypeBinding)b.right, 
                  (ParameterizedTypeBinding)bound.right);
              if (parameterizedTypeBinding != null) {
                b.right = (TypeBinding)parameterizedTypeBinding;
                return false;
              } 
            } 
          } 
        } 
        return this.subBounds.add(bound);
    } 
    throw new IllegalArgumentException("Unexpected bound relation in : " + bound);
  }
  
  public TypeBinding[] lowerBounds(boolean onlyProper, InferenceVariable variable) {
    TypeBinding[] boundTypes = new TypeBinding[this.superBounds.size()];
    Iterator<TypeBound> it = this.superBounds.iterator();
    long nullHints = variable.nullHints;
    int i = 0;
    while (it.hasNext()) {
      TypeBound current = it.next();
      TypeBinding boundType = current.right;
      if (!onlyProper || boundType.isProperType(true)) {
        boundTypes[i++] = boundType;
        nullHints |= current.nullHints;
      } 
    } 
    if (i == 0)
      return Binding.NO_TYPES; 
    if (i < boundTypes.length)
      System.arraycopy(boundTypes, 0, boundTypes = new TypeBinding[i], 0, i); 
    useNullHints(nullHints, boundTypes, variable.environment);
    InferenceContext18.sortTypes(boundTypes);
    return boundTypes;
  }
  
  public TypeBinding[] upperBounds(boolean onlyProper, InferenceVariable variable) {
    ReferenceBinding[] rights = new ReferenceBinding[this.subBounds.size()];
    TypeBinding simpleUpper = null;
    Iterator<TypeBound> it = this.subBounds.iterator();
    long nullHints = variable.nullHints;
    int i = 0;
    while (it.hasNext()) {
      TypeBinding right = ((TypeBound)it.next()).right;
      if (!onlyProper || right.isProperType(true)) {
        if (right instanceof ReferenceBinding) {
          rights[i++] = (ReferenceBinding)right;
          nullHints |= right.tagBits & 0x180000000000000L;
          continue;
        } 
        if (simpleUpper != null)
          return Binding.NO_TYPES; 
        simpleUpper = right;
      } 
    } 
    if (i == 0) {
      (new TypeBinding[1])[0] = simpleUpper;
      return (simpleUpper != null) ? new TypeBinding[1] : Binding.NO_TYPES;
    } 
    if (i == 1 && simpleUpper != null)
      return new TypeBinding[] { simpleUpper }; 
    if (i < rights.length)
      System.arraycopy(rights, 0, rights = new ReferenceBinding[i], 0, i); 
    useNullHints(nullHints, (TypeBinding[])rights, variable.environment);
    InferenceContext18.sortTypes((TypeBinding[])rights);
    return (TypeBinding[])rights;
  }
  
  public boolean hasDependency(InferenceVariable beta) {
    if (this.superBounds != null && hasDependency(this.superBounds, beta))
      return true; 
    if (this.sameBounds != null && hasDependency(this.sameBounds, beta))
      return true; 
    if (this.subBounds != null && hasDependency(this.subBounds, beta))
      return true; 
    if (this.inverseBounds != null && 
      this.inverseBounds.containsKey(beta))
      return true; 
    return false;
  }
  
  private boolean hasDependency(Set<TypeBound> someBounds, InferenceVariable var) {
    Iterator<TypeBound> bIt = someBounds.iterator();
    while (bIt.hasNext()) {
      TypeBound bound = bIt.next();
      if (TypeBinding.equalsEquals(bound.right, (TypeBinding)var) || bound.right.mentionsAny(new TypeBinding[] { (TypeBinding)var }, -1))
        return true; 
    } 
    return false;
  }
  
  public int size() {
    int size = 0;
    if (this.superBounds != null)
      size += this.superBounds.size(); 
    if (this.sameBounds != null)
      size += this.sameBounds.size(); 
    if (this.subBounds != null)
      size += this.subBounds.size(); 
    return size;
  }
  
  public int flattenInto(TypeBound[] collected, int idx) {
    if (this.superBounds != null) {
      int len = this.superBounds.size();
      System.arraycopy(this.superBounds.toArray(), 0, collected, idx, len);
      idx += len;
    } 
    if (this.sameBounds != null) {
      int len = this.sameBounds.size();
      System.arraycopy(this.sameBounds.toArray(), 0, collected, idx, len);
      idx += len;
    } 
    if (this.subBounds != null) {
      int len = this.subBounds.size();
      System.arraycopy(this.subBounds.toArray(), 0, collected, idx, len);
      idx += len;
    } 
    return idx;
  }
  
  public ThreeSets copy() {
    ThreeSets copy = new ThreeSets();
    if (this.superBounds != null)
      copy.superBounds = new HashSet<>(this.superBounds); 
    if (this.sameBounds != null)
      copy.sameBounds = new HashSet<>(this.sameBounds); 
    if (this.subBounds != null)
      copy.subBounds = new HashSet<>(this.subBounds); 
    copy.instantiation = this.instantiation;
    return copy;
  }
  
  public TypeBinding findSingleWrapperType() {
    if (this.instantiation != null && 
      this.instantiation.isProperType(true))
      switch (this.instantiation.id) {
        case 26:
        case 27:
        case 28:
        case 29:
        case 30:
        case 31:
        case 32:
        case 33:
          return this.instantiation;
      }  
    if (this.subBounds != null) {
      Iterator<TypeBound> it = this.subBounds.iterator();
      while (it.hasNext()) {
        TypeBinding boundType = ((TypeBound)it.next()).right;
        if (boundType.isProperType(true))
          switch (boundType.id) {
            case 26:
            case 27:
            case 28:
            case 29:
            case 30:
            case 31:
            case 32:
            case 33:
              return boundType;
          }  
      } 
    } 
    if (this.superBounds != null) {
      Iterator<TypeBound> it = this.superBounds.iterator();
      while (it.hasNext()) {
        TypeBinding boundType = ((TypeBound)it.next()).right;
        if (boundType.isProperType(true))
          switch (boundType.id) {
            case 26:
            case 27:
            case 28:
            case 29:
            case 30:
            case 31:
            case 32:
            case 33:
              return boundType;
          }  
      } 
    } 
    return null;
  }
  
  private void useNullHints(long nullHints, TypeBinding[] boundTypes, LookupEnvironment environment) {
    if (nullHints == 108086391056891904L) {
      for (int i = 0; i < boundTypes.length; i++)
        boundTypes[i] = boundTypes[i].withoutToplevelNullAnnotation(); 
    } else {
      AnnotationBinding[] annot = environment.nullAnnotationsFromTagBits(nullHints);
      if (annot != null)
        for (int i = 0; i < boundTypes.length; i++)
          boundTypes[i] = environment.createAnnotatedType(boundTypes[i], annot);  
    } 
  }
  
  TypeBinding combineAndUseNullHints(TypeBinding type, long nullHints, LookupEnvironment environment) {
    if (this.sameBounds != null) {
      Iterator<TypeBound> it = this.sameBounds.iterator();
      while (it.hasNext())
        nullHints |= ((TypeBound)it.next()).nullHints; 
    } 
    if (this.superBounds != null) {
      Iterator<TypeBound> it = this.superBounds.iterator();
      while (it.hasNext())
        nullHints |= ((TypeBound)it.next()).nullHints; 
    } 
    if (this.subBounds != null) {
      Iterator<TypeBound> it = this.subBounds.iterator();
      while (it.hasNext())
        nullHints |= ((TypeBound)it.next()).nullHints; 
    } 
    if (nullHints == 108086391056891904L)
      return type.withoutToplevelNullAnnotation(); 
    AnnotationBinding[] annot = environment.nullAnnotationsFromTagBits(nullHints);
    if (annot != null)
      return environment.createAnnotatedType(type, annot); 
    return type;
  }
  
  public void setInstantiation(TypeBinding type, InferenceVariable variable, LookupEnvironment environment) {
    if (environment.globalOptions.isAnnotationBasedNullAnalysisEnabled) {
      long variableBits = variable.tagBits & 0x180000000000000L;
      long allBits = type.tagBits | variableBits;
      if (this.instantiation != null)
        allBits |= this.instantiation.tagBits; 
      allBits &= 0x180000000000000L;
      if (allBits == 108086391056891904L)
        allBits = variableBits; 
      if (allBits != (type.tagBits & 0x180000000000000L)) {
        AnnotationBinding[] annot = environment.nullAnnotationsFromTagBits(allBits);
        if (annot != null) {
          type = environment.createAnnotatedType(type.withoutToplevelNullAnnotation(), annot);
        } else if (type.hasNullTypeAnnotations()) {
          type = type.withoutToplevelNullAnnotation();
        } 
      } 
    } 
    this.instantiation = type;
  }
}
